/**
 *  This file is part of MetaModel.
 *
 *  MetaModel is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  MetaModel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with MetaModel.  If not, see <http://www.gnu.org/licenses/>.
 */
package dk.eobjects.metamodel.detect;

import java.sql.Date;
import java.sql.Time;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.joda.time.DateTime;
import org.joda.time.LocalDate;
import org.joda.time.LocalTime;

import dk.eobjects.metamodel.schema.Column;
import dk.eobjects.metamodel.schema.ColumnType;
import dk.eobjects.metamodel.util.BooleanComparator;

/**
 * Class used to detect and modify the column type based on incoming string
 * values. The life-cycle of ColumnTypeTransformer objects are as follows:
 * <ul>
 * <li>Construct with the column</li>
 * <li>Execute the registerValue(String) for all (distinct) values of the
 * column.</li>
 * <li>Execute the narrowColumnType() method once which sets the appropriate
 * column type on the column.</li>
 * <li>Execute the narrowValue(String) for all values that are queried before
 * returning them to the user.</li>
 * </ul>
 * The column type transformer is currently able to detect the following column
 * types:
 * <ul>
 * <li>Strings (default if no other type is found)</li>
 * <li>Booleans</li>
 * <li>Integers</li>
 * <li>Doubles</li>
 * <li>Date</li>
 * <li>Time</li>
 * </ul>
 * And it will also detect if the column is nullable, ie. if null occurs in the
 * values.
 */
public class ColumnTypeTransformer {

	protected final Log _log = LogFactory.getLog(getClass());
	private boolean _booleanPossible = true;
	private boolean _integerPossible = true;
	private boolean _doublePossible = true;
	private boolean _datePossible = true;
	private boolean _timePossible = true;
	private boolean _nullPossible = false;
	private Column _column;

	public ColumnTypeTransformer(Column column) {
		if (column == null) {
			throw new IllegalArgumentException("Column cannot be null");
		}
		_column = column;
	}

	public Column getColumn() {
		return _column;
	}

	public void registerValue(String valueAsString) {
		if (valueAsString == null) {
			_nullPossible = true;
		} else {
			if (_booleanPossible) {
				try {
					BooleanComparator.parseBoolean(valueAsString);
				} catch (IllegalArgumentException e) {
					_booleanPossible = false;
				}
			}
			if (_doublePossible) {
				try {
					Double.parseDouble(valueAsString);
				} catch (NumberFormatException e) {
					_doublePossible = false;
					_integerPossible = false;
				}
				// If integer is possible, double will always also be possible,
				// but not nescesarily the other way around
				if (_integerPossible) {
					try {
						Integer.parseInt(valueAsString);
					} catch (NumberFormatException e) {
						_integerPossible = false;
					}
				}
			}
			if (_datePossible) {
				try {
					new LocalDate(valueAsString);
				} catch (IllegalArgumentException e) {
					_datePossible = false;
				}
			}
			if (_timePossible) {
				try {
					new LocalTime(valueAsString);
				} catch (IllegalArgumentException e) {
					_timePossible = false;
				}
			}
		}
	}

	public void narrowColumnType() {
		_log.debug("narrowColumnType()");
		_column.setNullable(_nullPossible);
		if (_booleanPossible) {
			_column.setType(ColumnType.BOOLEAN);
			_column.setNativeType("Text transformed to boolean");
		} else if (_integerPossible) {
			_column.setType(ColumnType.INTEGER);
			_column.setNativeType("Text transformed to integer");
		} else if (_doublePossible) {
			_column.setType(ColumnType.DOUBLE);
			_column.setNativeType("Text transformed to double");
		} else if (_datePossible) {
			_column.setType(ColumnType.DATE);
			_column.setNativeType("Text transformed to date");
		} else if (_timePossible) {
			_column.setType(ColumnType.TIME);
			_column.setNativeType("Text transformed to time");
		} else {
			_column.setType(ColumnType.VARCHAR);
			_column.setNativeType("Text");
		}
		if (_log.isInfoEnabled()) {
			_log.info("Column type narrowed in column '" + _column.getName()
					+ "': type=" + _column.getType() + ", nullable="
					+ _column.isNullable());
		}
	}

	public Object narrowValue(String valueAsString) {
		if (valueAsString != null) {
			ColumnType type = _column.getType();
			switch (type) {
			case VARCHAR:
				return valueAsString;
			case BOOLEAN:
				return BooleanComparator.parseBoolean(valueAsString);
			case INTEGER:
				return Integer.parseInt(valueAsString);
			case DOUBLE:
				return Double.parseDouble(valueAsString);
			case DATE:
				LocalDate localDate = new LocalDate(valueAsString);
				return new Date(new DateTime(localDate.getYear(), localDate
						.getMonthOfYear(), localDate.getDayOfMonth(), 0, 0, 0,
						0).getMillis());
			case TIME:
				LocalTime localTime = new LocalTime(valueAsString);
				return new Time(new DateTime(1970, 1, 1, localTime
						.getHourOfDay(), localTime.getMinuteOfHour(), localTime
						.getSecondOfMinute(), localTime.getMillisOfSecond())
						.getMillis());
			default:
				throw new IllegalStateException(
						"Encountered unsupported column type: " + type);
			}
		}
		return null;
	}
}